/* +------------------------------------------------------------------------+
   |                                                                        |
   |                  Entiers de longueur arbitraire                        |
   |                                                                        |
   |                       Allocation mmoire                               |
   |                                                                        |
   +------------------------------------------------------------------------+ */

/* M. Quercia, 10/08/2001 */

/*
 Cont(x)
    entier point par la rfrence x

 Val(x)
    transtypage (entier *) -> value

 Alloc(x,l)
    alloue l'entier x pour contenir l chiffres

 Alloc_x_y( a1,l1, ..., ax,lx,  b1, ..., by )
    alloue les entiers a1,..,xy pour contenir l1,..,lx chiffres
    b1,..,by pointent vers des valeurs dj alloues risquant
    d'tre dplaces par le GC.

 AllocN_x_y( v,n, a1,l1, ..., ax,lx, b1, ..., by )
    idem mais alloue en plus le vecteur v de longueur n, v est allou
    en dernier donc on peut assigner ses champs directement.

 Enlarge_x_y( a1,b1,l1 ...,ax,bx,lx, c1, .., cy )
    -- vrifie que les entiers rfrencs par b1,..,bx sont assez grands
    pour recevoir l1,..,lx chiffres et ralloue ceux qui sont trop
    petits (en prenant un chiffre de marge pour limiter les rallocations
    successives).
    -- initialise ensuite les pointeurs ai sur les entiers anciens ou
    rallous rfrencs par les bi.
    -- prserve les liaisons aux valeurs pointes par c1,...,cy en cas
    de rorganisation mmoire par le GC
*/


#define Cont(a)     ((entier *)Field(a,0))
#define Val(a)      *(value *)&a
#define Capacity(x) (C_of_W(Wosize_val(Cont(x))-1))

#ifndef use_camllight
#if OCAML_VERSION < 202
#define use_camllight
#endif
#endif

#ifndef use_camllight

                    /* +----------------------+
                       |  Pour Ocaml >= 1.07  |
                       +----------------------+ */


#define Alloc(x,l) {                                    \
  unsigned long alloc_size = W_of_C(l)+1;               \
  if (alloc_size > Max_wosize)                          \
    failwith("create too big a number");                \
   x = (entier *) (alloc(alloc_size,Abstract_tag));     \
}


#define Alloc_1_1(a,l,b) {               \
  Begin_roots1(Val(b));                  \
  Alloc(a,l);                            \
  End_roots();                           \
}				         

#define Alloc_1_1t(a,l,ty,b) {           \
  Begin_roots1(Val(b));                  \
  Alloc(a,l);                            \
  End_roots();                           \
}				         

#define Alloc_1_2(a,l,b,c) {             \
  Begin_roots2(Val(b),Val(c));           \
  Alloc(a,l);                            \
  End_roots();                           \
}				         

#define Alloc_1_3(a,l,b,c,d) {           \
  Begin_roots3(Val(b),Val(c),Val(d));    \
  Alloc(a,l);                            \
  End_roots();                           \
}				         

#define Alloc_2_1(a,la,b,lb,c) {         \
  a = (entier *)Val_unit;		 \
  b = (entier *)Val_unit;		 \
  Begin_roots3(Val(a),Val(b),Val(c));    \
  Alloc(a,la);				 \
  Alloc(b,lb);				 \
  End_roots();			         \
}

#define AllocN_1_1(v,n,a,l,b) {          \
  a = (entier *)Val_unit;		 \
  Begin_roots2(Val(a),Val(b));		 \
  Alloc(a,l);				 \
  v = alloc_tuple(n);			 \
  End_roots();			         \
}

#define AllocN_2_1(v,n,a,la,b,lb,c) {    \
  a = (entier *)Val_unit;		 \
  b = (entier *)Val_unit;		 \
  Begin_roots3(Val(a),Val(b),Val(c));    \
  Alloc(a,la);				 \
  Alloc(b,lb);				 \
  v = alloc_tuple(n);			 \
  End_roots();			         \
}

#define AllocN_2_2(v,n,a,la,b,lb,c,d) {  \
  a = (entier *)Val_unit;		 \
  b = (entier *)Val_unit;		 \
  Begin_roots4(Val(a),Val(b),Val(c),Val(d)); \
  Alloc(a,la);				 \
  Alloc(b,lb);				 \
  v = alloc_tuple(n);			 \
  End_roots();			         \
}

#define AllocN_3_2(v,n,a,la,b,lb,c,lc,d,e) {  \
  a = (entier *)Val_unit;		 \
  b = (entier *)Val_unit;		 \
  c = (entier *)Val_unit;		 \
  Begin_roots5(Val(a),Val(b),Val(c),Val(d),Val(e)); \
  Alloc(a,la);				 \
  Alloc(b,lb);				 \
  Alloc(c,lc);				 \
  v = alloc_tuple(n);			 \
  End_roots();		                 \
}

#define AllocN_5_2(v,n,a,la,b,lb,c,lc,d,ld,e,le,f,g) {  \
  a = (entier *)Val_unit;		 \
  b = (entier *)Val_unit;		 \
  c = (entier *)Val_unit;		 \
  d = (entier *)Val_unit;		 \
  e = (entier *)Val_unit;		 \
  Begin_roots5(Val(a),Val(b),Val(c),Val(d),Val(e)); \
  Begin_roots2(Val(f),Val(g));           \
  Alloc(a,la);				 \
  Alloc(b,lb);				 \
  Alloc(c,lc);				 \
  Alloc(d,ld);				 \
  Alloc(e,le);				 \
  v = alloc_tuple(n);			 \
  End_roots();			         \
  End_roots();		                 \
}

#define Enlarge(a,x,l) {                 \
  if (Capacity(x) < (l)) {               \
    Begin_roots1(x);                     \
    Alloc(a,2*(l)+1);		         \
    modify(&Field(x,0),(value) a);       \
    End_roots();		         \
  } else a = Cont(x);                    \
}				         

#define Enlarge_1_1(a,x,l,b) {           \
  if (Capacity(x) < (l)) {               \
    Begin_roots2(x, Val(b));             \
    Alloc(a,2*(l)+1);		         \
    modify(&Field(x,0),(value) a);       \
    End_roots();		         \
  } else a = Cont(x);                    \
}				         

#define Enlarge_1_2(a,x,l,b,c) {         \
  if (Capacity(x) < (l)) {               \
    Begin_roots3(x, Val(b), Val(c));     \
    Alloc(a,2*(l)+1);		         \
    modify(&Field(x,0),(value) a);       \
    End_roots();		         \
  } else a = Cont(x);                    \
}

#define Enlarge_1_3(a,x,l,b,c,d) {       \
  if (Capacity(x) < (l)) {               \
    Begin_roots4(x, Val(b),Val(c),Val(d)); \
    Alloc(a,2*(l)+1);		         \
    modify(&Field(x,0),(value) a);       \
    End_roots();		         \
  } else a = Cont(x);                    \
}

#define Enlarge_2_1(a,x,la,b,y,lb,c) {             \
  int small_x = (Capacity(x) < (la));              \
  int small_y = (Capacity(y) < (lb));              \
  if (small_x || small_y) {			   \
    Begin_roots3(x,y, Val(c));		           \
    if (small_x) {				   \
      Alloc(a,2*(la)+1);			   \
      modify(&Field(x,0),(value) a);		   \
    } else a = Cont(x);				   \
    if (small_y) {				   \
      Alloc(b,2*(lb)+1);			   \
      modify(&Field(y,0),(value) b);		   \
    } else b = Cont(y);				   \
    End_roots();                                   \
  } else {a = Cont(x); b = Cont(y);}               \
}


#define Enlarge_2_2(a,x,la,b,y,lb,c,d) {           \
  int small_x = (Capacity(x) < (la));              \
  int small_y = (Capacity(y) < (lb));              \
  if (small_x || small_y) {			   \
    Begin_roots4(x,y, Val(c), Val(d));		   \
    if (small_x) {				   \
      Alloc(a,2*(la)+1);			   \
      modify(&Field(x,0),(value) a);		   \
    } else a = Cont(x);				   \
    if (small_y) {				   \
      Alloc(b,2*(lb)+1);			   \
      modify(&Field(y,0),(value) b);		   \
    } else b = Cont(y);				   \
    End_roots();                                   \
  } else {a = Cont(x); b = Cont(y);}               \
}

#define Enlarge_3_2(c,x,lc,d,y,ld,e,z,le,f,g) {    \
  int small_x = (Capacity(x) < (lc));              \
  int small_y = (Capacity(y) < (ld));              \
  int small_z = (Capacity(z) < (le));              \
  if (small_x || small_y || small_z) {             \
    Begin_roots5(x,y,z,Val(f),Val(g));             \
    if (small_x) {				   \
      Alloc(c,2*(lc)+1);			   \
      modify(&Field(x,0),(value) c);		   \
    } else c = Cont(x);				   \
    if (small_y) {				   \
      Alloc(d,2*(ld)+1);			   \
      modify(&Field(y,0),(value) d);		   \
    } else d = Cont(y);				   \
    if (small_z) {				   \
      Alloc(e,2*(le)+1);			   \
      modify(&Field(z,0),(value) e);		   \
    } else e = Cont(z);				   \
    End_roots();                                   \
  } else {c = Cont(x); d = Cont(y); e = Cont(z);}  \
}

#define Enlarge_5_2(a,r,la,b,s,lb,c,x,lc,d,y,ld,e,z,le,f,g) { \
  int small_r = (Capacity(r) < (la));              \
  int small_s = (Capacity(s) < (lb));              \
  int small_x = (Capacity(x) < (lc));              \
  int small_y = (Capacity(y) < (ld));              \
  int small_z = (Capacity(z) < (le));              \
  if (small_r || small_s || small_x || small_y || small_z) {  \
    Begin_roots5(r,s,x,y,z);		           \
    Begin_roots2(Val(f), Val(g));		   \
    if (small_r) {				   \
      Alloc(a,2*(la)+1);			   \
      modify(&Field(r,0),(value) a);		   \
    } else a = Cont(r);				   \
    if (small_s) {				   \
      Alloc(b,2*(lb)+1);			   \
      modify(&Field(s,0),(value) b);		   \
    } else b = Cont(s);				   \
    if (small_x) {				   \
      Alloc(c,2*(lc)+1);			   \
      modify(&Field(x,0),(value) c);		   \
    } else c = Cont(x);				   \
    if (small_y) {				   \
      Alloc(d,2*(ld)+1);			   \
      modify(&Field(y,0),(value) d);		   \
    } else d = Cont(y);				   \
    if (small_z) {				   \
      Alloc(e,2*(le)+1);			   \
      modify(&Field(z,0),(value) e);		   \
    } else e = Cont(z);				   \
    End_roots();                                   \
    End_roots();                                   \
  } else {a = Cont(r); b = Cont(s); c = Cont(x); d = Cont(y); e = Cont(z);} \
}

#else /* #ifndef use_camllight */

                 /* +-----------------------------+
                    |  Camllight et Ocaml < 1.07  |
                    +-----------------------------+ */


#define Alloc(x,l) {                                    \
  unsigned long alloc_size = W_of_C(l)+1;               \
  if (alloc_size > Max_wosize)                          \
    failwith("create too big a number");                \
   x = (entier *) ((alloc_size <= Max_young_wosize) ?   \
                  alloc    (alloc_size,Abstract_tag) :  \
                  alloc_shr(alloc_size,Abstract_tag));  \
  }

#define Allocv(x,l) {                                   \
  unsigned long alloc_size = W_of_C(l)+1;               \
  if (alloc_size > Max_wosize)                          \
    failwith("create too big a number");                \
   x = ((alloc_size <= Max_young_wosize) ?              \
                  alloc    (alloc_size,Abstract_tag) :  \
                  alloc_shr(alloc_size,Abstract_tag));  \
  }

#define Alloc_1_1(a,l,b) {               \
  Push_roots(__v,1);                     \
  __v[0] = (value)b;                     \
  Alloc(a,l);                            \
  b = (entier *)__v[0];                  \
  Pop_roots();                           \
}				         

#define Alloc_1_1t(a,l,ty,b) {           \
  Push_roots(__v,1);                     \
  __v[0] = (value)b;                     \
  Alloc(a,l);                            \
  b = (ty)__v[0];                        \
  Pop_roots();                           \
}				         

#define Alloc_1_2(a,l,b,c) {             \
  Push_roots(__v,2);                     \
  __v[0] = (value)b;                     \
  __v[1] = (value)c;                     \
  Alloc(a,l);                            \
  b = (entier *)__v[0];                  \
  c = (entier *)__v[1];                  \
  Pop_roots();                           \
}				         

#define Alloc_1_3(a,l,b,c,d) {           \
  Push_roots(__v,3);                     \
  __v[0] = (value)b;                     \
  __v[1] = (value)c;                     \
  __v[2] = (value)d;                     \
  Alloc(a,l);                            \
  b = (entier *)__v[0];                  \
  c = (entier *)__v[1];                  \
  d = (entier *)__v[2];                  \
  Pop_roots();                           \
}				         

#define Alloc_2_1(a,la,b,lb,c) {         \
  Push_roots(__v,2);                     \
  __v[0] = (value)c;                     \
  Allocv(__v[1],la);                     \
  Alloc(b,lb);                           \
  a = (entier *)__v[1];                  \
  c = (entier *)__v[0];                  \
  Pop_roots();                           \
}				         

#define AllocN_1_1(v,n,a,l,b) {          \
  Push_roots(__v,2);                     \
  __v[0] = (value)b;                     \
  Allocv(__v[1],l);                      \
  v = alloc_tuple(n);			 \
  a = (entier *)__v[1];                  \
  b = (entier *)__v[0];                  \
  Pop_roots();                           \
}				         

#define AllocN_2_1(v,n,a,la,b,lb,c) {    \
  Push_roots(__v,3);                     \
  __v[0] = (value)c;                     \
  Allocv(__v[1],la);                     \
  Allocv(__v[2],lb);                     \
  v = alloc_tuple(n);			 \
  a = (entier *)__v[1];                  \
  b = (entier *)__v[2];                  \
  c = (entier *)__v[0];                  \
  Pop_roots();                           \
}				         

#define AllocN_2_2(v,n,a,la,b,lb,c,d) {  \
  Push_roots(__v,4);                     \
  __v[0] = (value)c;                     \
  __v[1] = (value)d;                     \
  Allocv(__v[2],la);                     \
  Allocv(__v[3],lb);                     \
  v = alloc_tuple(n);			 \
  a = (entier *)__v[2];                  \
  b = (entier *)__v[3];                  \
  c = (entier *)__v[0];                  \
  d = (entier *)__v[1];                  \
  Pop_roots();                           \
}				         

#define AllocN_3_2(v,n,a,la,b,lb,c,lc,d,e) {  \
  Push_roots(__v,5);                     \
  __v[0] = (value)d;                     \
  __v[1] = (value)e;                     \
  Allocv(__v[2],la);                     \
  Allocv(__v[3],lb);                     \
  Allocv(__v[4],lc);                     \
  v = alloc_tuple(n);			 \
  a = (entier *)__v[2];                  \
  b = (entier *)__v[3];                  \
  c = (entier *)__v[4];                  \
  d = (entier *)__v[0];                  \
  e = (entier *)__v[1];                  \
  Pop_roots();                           \
}				         

#define AllocN_5_2(v,n,a,la,b,lb,c,lc,d,ld,e,le,f,g) {  \
  Push_roots(__v,7);                     \
  __v[0] = (value)f;                     \
  __v[1] = (value)g;                     \
  Allocv(__v[2],la);                     \
  Allocv(__v[3],lb);                     \
  Allocv(__v[4],lc);                     \
  Allocv(__v[5],ld);                     \
  Allocv(__v[6],le);                     \
  v = alloc_tuple(n);			 \
  a = (entier *)__v[2];                  \
  b = (entier *)__v[3];                  \
  c = (entier *)__v[4];                  \
  d = (entier *)__v[5];                  \
  e = (entier *)__v[6];                  \
  f = (entier *)__v[0];                  \
  g = (entier *)__v[1];                  \
  Pop_roots();                           \
}				         

#define Enlarge(a,x,l) {                 \
  if (Capacity(x) < (l)) {               \
    Push_roots(__v,1);                   \
    __v[0] = (value)x;                   \
    Alloc(a,2*(l)+1);		         \
    x = __v[0];                          \
    modify(&Field(x,0),(value) a);       \
    Pop_roots();		         \
  } else a = Cont(x);                    \
}				         

#define Enlarge_1_1(a,x,l,b) {           \
  if (Capacity(x) < (l)) {               \
    Push_roots(__v,2);                   \
    __v[0] = (value)x;                   \
    __v[1] = (value)b;                   \
    Alloc(a,2*(l)+1);		         \
    x = __v[0];                          \
    b = (entier *)__v[1];                \
    modify(&Field(x,0),(value) a);       \
    Pop_roots();		         \
  } else a = Cont(x);                    \
}				         

#define Enlarge_1_2(a,x,l,b,c) {         \
  if (Capacity(x) < (l)) {               \
    Push_roots(__v,3);                   \
    __v[0] = (value)x;                   \
    __v[1] = (value)b;                   \
    __v[2] = (value)c;                   \
    Alloc(a,2*(l)+1);		         \
    x = __v[0];                          \
    b = (entier *)__v[1];                \
    c = (entier *)__v[2];                \
    modify(&Field(x,0),(value) a);       \
    Pop_roots();		         \
  } else a = Cont(x);                    \
}				         

#define Enlarge_1_3(a,x,l,b,c,d) {       \
  if (Capacity(x) < (l)) {               \
    Push_roots(__v,4);                   \
    __v[0] = (value)x;                   \
    __v[1] = (value)b;                   \
    __v[2] = (value)c;                   \
    __v[3] = (value)d;                   \
    Alloc(a,2*(l)+1);		         \
    x = __v[0];                          \
    b = (entier *)__v[1];                \
    c = (entier *)__v[2];                \
    d = (entier *)__v[3];                \
    modify(&Field(x,0),(value) a);       \
    Pop_roots();		         \
  } else a = Cont(x);                    \
}				         

#define Enlarge_2_1(a,x,la,b,y,lb,c) {   \
  int small_x = (Capacity(x) < (la));    \
  int small_y = (Capacity(y) < (lb));    \
  if (small_x || small_y) {		 \
    Push_roots(__v,4);                   \
    __v[0] = (value)x;                   \
    __v[1] = (value)y;                   \
    __v[2] = (value)c;                   \
    if (small_x) {                       \
      Allocv(__v[3],2*(la)+1);	         \
      modify(&Field(__v[0],0),__v[3]);   \
    } else __v[3] = Field(__v[0],0);     \
    if (small_y) {                       \
      Alloc(b,2*(lb)+1);		 \
      modify(&Field(__v[1],0),(value)b); \
    } else b = Cont(__v[1]);             \
    x = __v[0];                          \
    y = __v[1];                          \
    c = (entier *)__v[2];                \
    a = (entier *)__v[3];                \
    Pop_roots();		         \
  } else {a = Cont(x); b = Cont(y);}     \
}				         

#define Enlarge_2_2(a,x,la,b,y,lb,c,d) { \
  int small_x = (Capacity(x) < (la));    \
  int small_y = (Capacity(y) < (lb));    \
  if (small_x || small_y) {		 \
    Push_roots(__v,5);                   \
    __v[0] = (value)x;                   \
    __v[1] = (value)y;                   \
    __v[2] = (value)c;                   \
    __v[3] = (value)d;                   \
    if (small_x) {                       \
      Allocv(__v[4],2*(la)+1);	         \
      modify(&Field(__v[0],0),__v[4]);   \
    } else __v[4] = Field(__v[0],0);     \
    if (small_y) {                       \
      Alloc(b,2*(lb)+1);		 \
      modify(&Field(__v[1],0),(value)b); \
    } else b = Cont(__v[1]);             \
    x = __v[0];                          \
    y = __v[1];                          \
    c = (entier *)__v[2];                \
    d = (entier *)__v[3];                \
    a = (entier *)__v[4];                \
    Pop_roots();		         \
  } else {a = Cont(x); b = Cont(y);}     \
}				         

#define Enlarge_3_2(a,x,la,b,y,lb,c,z,lc,d,e) { \
  int small_x = (Capacity(x) < (la));    \
  int small_y = (Capacity(y) < (lb));    \
  int small_z = (Capacity(z) < (lc));    \
  if (small_x || small_y || small_z) {	 \
    Push_roots(__v,7);                   \
    __v[0] = (value)x;                   \
    __v[1] = (value)y;                   \
    __v[2] = (value)z;                   \
    __v[3] = (value)d;                   \
    __v[4] = (value)e;                   \
    if (small_x) {                       \
      Allocv(__v[5],2*(la)+1);           \
      modify(&Field(__v[0],0),__v[5]);   \
    } else __v[5] = Field(__v[0],0);     \
    if (small_y) {                       \
      Allocv(__v[6],2*(lb)+1);	         \
      modify(&Field(__v[1],0),__v[6]);   \
    } else __v[6] = Field(__v[1],0);     \
    if (small_z) {                       \
      Alloc(c,2*(lc)+1);		 \
      modify(&Field(__v[2],0),(value)c); \
    } else c = Cont(__v[2]);             \
    x = __v[0];                          \
    y = __v[1];                          \
    z = __v[2];                          \
    d = (entier *)__v[3];                \
    e = (entier *)__v[4];                \
    a = (entier *)__v[5];                \
    b = (entier *)__v[6];                \
    Pop_roots();		         \
  } else {a = Cont(x); b = Cont(y); c = Cont(z); }     \
}				         

#define Enlarge_5_2(a,x,la,b,y,lb,c,z,lc,d,r,ld,e,s,le,f,g) { \
  int small_x = (Capacity(x) < (la));    \
  int small_y = (Capacity(y) < (lb));    \
  int small_z = (Capacity(z) < (lc));    \
  int small_r = (Capacity(r) < (ld));    \
  int small_s = (Capacity(s) < (le));    \
  if (small_x || small_y || small_z || small_r || small_s) {	 \
    Push_roots(__v,11);                  \
    __v[0] = (value)x;                   \
    __v[1] = (value)y;                   \
    __v[2] = (value)z;                   \
    __v[3] = (value)r;                   \
    __v[4] = (value)s;                   \
    __v[5] = (value)f;                   \
    __v[6] = (value)g;                   \
    if (small_x) {                       \
      Allocv(__v[7],2*(la)+1);	         \
      modify(&Field(__v[0],0),__v[7]);   \
    } else __v[7] = Field(__v[0],0);     \
    if (small_y) {                       \
      Allocv(__v[8],2*(lb)+1);	         \
      modify(&Field(__v[1],0),__v[8]);   \
    } else __v[8] = Field(__v[1],0);     \
    if (small_z) {                       \
      Allocv(__v[9],2*(lc)+1);	         \
      modify(&Field(__v[2],0),__v[9]);   \
    } else __v[9] = Field(__v[2],0);     \
    if (small_r) {                       \
      Allocv(__v[10],2*(ld)+1);          \
      modify(&Field(__v[3],0),__v[10]);  \
    } else __v[10] = Field(__v[3],0);    \
    if (small_s) {                       \
      Alloc(e,2*(le)+1);		 \
      modify(&Field(__v[4],0),(value)e); \
    } else e = Cont(__v[4]);             \
    x = __v[0];                          \
    y = __v[1];                          \
    z = __v[2];                          \
    r = __v[3];                          \
    s = __v[4];                          \
    f = (entier *)__v[5];                \
    g = (entier *)__v[6];                \
    a = (entier *)__v[7];                \
    b = (entier *)__v[8];                \
    c = (entier *)__v[9];                \
    d = (entier *)__v[10];               \
    Pop_roots();		         \
  } else {a = Cont(x); b = Cont(y); c = Cont(z); d = Cont(r); e = Cont(s);} \
}				         

#endif /* #ifndef use_camllight */

