/* +------------------------------------------------------------------------+
   |                                                                        |
   |                  Entiers de longueur arbitraire                        |
   |                                                                        |
   |                       Interface C <-> Ocaml                            |
   |                                                                        |
   +------------------------------------------------------------------------+ */

/* M. Quercia, 10/08/2001 */

#include <string.h>
#include <stdio.h>
#include <ctype.h>

#ifdef use_camllight

/* Camllight */
#include <mlvalues.h>
#include <alloc.h>
#include <memory.h>
#define add_fail

#else

/* Ocaml toutes versions */
#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>

#if OCAML_VERSION >= 202
#include <caml/fail.h>
#else
#define add_fail
#endif
#endif /* #ifdef use_camllight */

#ifdef add_fail
#ifdef __GNUC__
/* Works only in GCC 2.5 and later */
#define Noreturn __attribute ((noreturn))
#else
#define Noreturn
#endif

void failwith (char *) Noreturn;
#endif /* #ifdef add_fail */

#include "ml-alloc.h"
#include "ml-long_int.h"


             /* +------------------------------------+
                |  Gestion des oprations sur place  |
                +------------------------------------+ */

/*
  copy_in(x,b) : x <- b
  copy_out(a)  : a     
*/

void xx(copy_in)(value x, entier *b) {
  entier *a;
  longueur l = Lg(b);
  Enlarge_1_1(a,x,l,b);
  memmove(a, b, l*sizeof(chiffre)+sizeof(longueur));
}

value xx(copy_out)(value a) {
  longueur la = Lg(Cont(a));
  entier *b;
  Alloc_1_1t(b,la,value,a);
  memcpy(b, Cont(a), la*sizeof(chiffre)+sizeof(longueur));
  return((value) b);
}

             /* +-------------------------------------+
                |  Accs  la reprsentation binaire  |
                +-------------------------------------+ */

/*
  nbits(a)      : taille de a en base 2
  lowbits(a)    : bits 0..31 de |a|
  highbits(a)   : bits l-31..l-1 de |a|, l = nbits(a)
  nth_word(a,n) : bits 16*(n-1)..16*n-1 de |a|
  of_int(a)     : a
  int_of(a)     : a
*/

value xx(nbits)(entier *a) {
  chiffre  x;
  longueur l = Lg(a);
  if (l) for (l--, x = a->val[l], l *= HW; (x); x >>= 1, l++);
  return(Val_long(l));
}

value xx(lowbits)(entier *a) {
  longueur la = Lg(a);
  ndouble  x = (la == 0) ? 0
             : (la == 1) ? a->val[0]
             : ((ndouble)a->val[1] << HW) + (ndouble)a->val[0];
  return(Val_long(x & 0x7fffffff));
}

value xx(highbits)(entier *a) {
  longueur la = Lg(a);
  ndouble  x,y;

  if (la == 0) return(Val_long(0));

  x = (ndouble)a->val[la-1] << HW;
  if (la >= 2) x += (ndouble)a->val[la-2];
  y = (la >= 3) ? a->val[la-3] : 0;
  while (x < (ndouble)0x40000000) {
    x <<= 1;
    y <<= 1;
    if (y >= BASE) {x++; y -= BASE;}
  }
  while (x >= (ndouble)0x80000000) x >>= 1;
  return(Val_long(x));
}

value xx(nth_word)(entier *a, value n) {
  longueur la = Lg(a);
  longueur nn = Long_val(n), n1,n2;

  if (nn < 0) failwith("negative word count");
  n1 = nn/(sizeof(chiffre)/2);
  n2 = 16*(nn%(sizeof(chiffre)/2));
  return(Val_long((n1 < la) ? (a->val[n1] >> n2) & 0xffff : 0));
}

value xx(of_int)(value a) {
  entier *c;
  zdouble  aa = Long_val(a);
  longueur sa = (aa < 0) ? SIGN_m : 0;
  if (sa) aa = -aa;

  if (aa == 0)        {Alloc(c,0); c->hd = 0;}
  else if (aa < BASE) {Alloc(c,1); c->hd = 1|sa; c->val[0] = aa;}
  else                {Alloc(c,2); c->hd = 2|sa; c->val[0] = aa; c->val[1] = aa >> HW;}

  return((value) c);
}

value xx(int_of)(entier *a) {
  longueur la = Lg(a);
  ndouble aa = (la == 0) ? 0
             : (la == 1) ? a->val[0]
             : ((ndouble)a->val[1] << HW) + (ndouble)a->val[0];
  if ((la > 2) | (aa > 0x7FFFFFFF)) failwith ("integer overflow");
  return(Val_long((Signe(a)) ? -aa : aa));
}


                            /* +----------------+
                               |  Comparaisons  |
                               +----------------+ */

value xx(eq)   (entier *a, entier *b) {return((xz(cmp)(a,b) == 0) ? Val_true  : Val_false);}
value xx(neq)  (entier *a, entier *b) {return((xz(cmp)(a,b) != 0) ? Val_true  : Val_false);}
value xx(inf)  (entier *a, entier *b) {return((xz(cmp)(a,b) <  0) ? Val_true  : Val_false);}
value xx(infeq)(entier *a, entier *b) {return((xz(cmp)(a,b) <= 0) ? Val_true  : Val_false);}
value xx(sup)  (entier *a, entier *b) {return((xz(cmp)(a,b) >  0) ? Val_true  : Val_false);}
value xx(supeq)(entier *a, entier *b) {return((xz(cmp)(a,b) >= 0) ? Val_true  : Val_false);}
value xx(cmp)  (entier *a, entier *b) {return(Val_long(xz(cmp)(a,b)));}
value xx(sgn)  (entier *a, entier *b) {return(Val_long((Lg(a)) ? ((Signe(a)) ? -1 : 1) : 0));}


                       /* +--------------------------+
                          |  Valeur absolue, oppos  |
                          +--------------------------+ */

value xx(abs)(entier *a) {
  longueur la = Lg(a);
  entier *c;
  Alloc_1_1(c,la,a);
  memcpy(c->val,a->val,la*sizeof(chiffre));
  c->hd = la;
  return((value) c);
}

void xx(abs_in)(value x, entier *a) {
  longueur la = Lg(a);
  entier *c;
  Enlarge_1_1(c,x,la,a);
  if (c != a) memcpy(c->val,a->val,la*sizeof(chiffre));
  c->hd = la;
}

value xx(neg)(entier *a) {
  longueur la = Lg(a);
  entier *c;
  Alloc_1_1(c,la,a);
  memcpy(c->val,a->val,la*sizeof(chiffre));
  c->hd = (la) ? (la | (Signe(a)^SIGN_m)) : 0;
  return((value) c);
}

void xx(neg_in)(value x, entier *a) {
  longueur la = Lg(a);
  entier *c;
  Enlarge_1_1(c,x,la,a);
  if (c != a) memcpy(c->val,a->val,la*sizeof(chiffre));
  c->hd = (la) ? (la | (Signe(a)^SIGN_m)) : 0;
}


                        /* +-------------------------+
                           |  Addition/soustraction  |
                           +-------------------------+ */

value xx(add)(entier *a, entier *b) {
  entier *c;
  longueur la = Lg(a), lb = Lg(b);
  Alloc_1_2(c,max(la,lb)+1,a,b);
  xz(add)(a,b,c);
  return((value) c);
}

value xx(sub)(entier *a, entier *b) {
  entier *c;
  longueur la = Lg(a), lb = Lg(b);
  Alloc_1_2(c,max(la,lb)+1,a,b);
  xz(sub)(a,b,c);
  return((value) c);
}

void xx(add_in)(value x, entier *a, entier *b) {
  entier *c;
  longueur la = Lg(a), lb = Lg(b);
  Enlarge_1_2(c,x,max(la,lb)+1,a,b);
  xz(add)(a,b,c);
}

void xx(sub_in)(value x, entier *a, entier *b) {
  entier *c;
  longueur la = Lg(a), lb = Lg(b);
  Enlarge_1_2(c,x,max(la,lb)+1,a,b);
  xz(sub)(a,b,c);
}


                              /* +-------------+
                                 |  Dcalages  |
                                 +-------------+ */


value xx(shl)(entier *a, value b) {
  entier *c;
  longueur la = Lg(a), lc;
  longueur bb = Long_val(b);

  lc = (bb >= 0) ? la + bb/HW + 1 : la - (-bb/HW);
  if (lc > 0) {
    Alloc_1_1(c,lc,a);
    xz(shift)(a, bb, c);
  } else {
    Alloc(c,0);
    c->hd = 0;
  }
  return((value) c);
}

value xx(shr)(entier *a, value b) {
  entier *c;
  longueur la = Lg(a), lc;
  longueur bb = -Long_val(b);

  lc = (bb >= 0) ? la + bb/HW + 1 : la - (-bb/HW);
  if (lc > 0) {
    Alloc_1_1(c,lc,a);
    xz(shift)(a, bb, c);
  } else {
    Alloc(c,0);
    c->hd = 0;
  }
  return((value) c);
}

void xx(shl_in)(value x, entier *a, value b) {
  entier *c;
  longueur la = Lg(a), lc;
  longueur bb = Long_val(b);

  lc = (bb >= 0) ? la + bb/HW + 1 : la - (-bb/HW);
  if (lc > 0) {
    Enlarge_1_1(c,x,lc,a);
    xz(shift)(a, bb, c);
  } else {
    Cont(x)->hd = 0;
  }
}

void xx(shr_in)(value x, entier *a, value b) {
  entier *c;
  longueur la = Lg(a), lc;
  longueur bb = -Long_val(b);

  lc = (bb >= 0) ? la + bb/HW + 1 : la - (-bb/HW);
  if (lc > 0) {
    Enlarge_1_1(c,x,lc,a);
    xz(shift)(a, bb, c);
  } else {
    Cont(x)->hd = 0;
  }
}


value xx(split)(entier *a, value b) {
  entier *c,*d;
  longueur la = Lg(a);
  longueur bb = Long_val(b);
  longueur lc = max(la-bb/HW+1,0);
  longueur ld = min(la,bb/HW+1);
  value res;

  if (bb < 0) failwith("negative split index");
  AllocN_2_1(res,2,c,lc,d,ld,a);
  xz(split)(a,bb,c,d);
  Field(res,0) = (value)c;
  Field(res,1) = (value)d;
  return(res);
}

void xx(split_in)(value x, value y, entier *a, value b) {
  entier *c,*d;
  longueur la = Lg(a);
  longueur bb = Long_val(b);
  longueur lc = max(la-bb/HW,0);
  longueur ld = min(la,(bb+HW-1)/HW);

  if (bb < 0) failwith("negative split index");
  if (x == y)  failwith("remainder and quotient share the same memory");
  Enlarge_2_1(c,x,lc,d,y,ld,a);
  xz(split)(a,bb,c,d);
}

value xx(join)(entier *a, entier *b, value c) {
  long     cc = Long_val(c);
  entier   *d;
  longueur la = Lg(a), lb = Lg(b), ld;

  if (cc < 0) failwith("negative join index");
  ld = max(la,lb+cc/HW+1)+1;
  Alloc_1_2(d,ld,a,b);
  xz(join)(a,b,cc,d);
  return((value) d);
}

void xx(join_in)(value x,entier *a, entier *b, value c) {
  long     cc = Long_val(c);
  entier   *d;
  longueur la = Lg(a), lb = Lg(b), ld;

  if (cc < 0) failwith("negative join index");
  ld = max(la,lb+cc/HW+1)+1;
  Enlarge_1_2(d,x,ld,a,b);
  xz(join)(a,b,cc,d);
}

                           /* +------------------+
                              |  Multiplication  |
                              +------------------+ */

value xx(mul_1)(entier *a, value b) {
  entier *c;
  longueur la = Lg(a);
  long bb = Long_val(b);

  Alloc_1_1(c,la+2,a);
  xz(mul_2)(a, bb, c);
  return((value) c);
}    

void xx(mul_1_in)(value x, entier *a, value b) {
  entier *c;
  longueur la = Lg(a);
  long bb = Long_val(b);

  Enlarge_1_1(c,x,la+2,a);
  xz(mul_2)(a, bb, c);
}    

value xx(mul_k)(entier *a, entier *b) {
  entier *c;
  longueur la = Lg(a), lb = Lg(b);
  Alloc_1_2(c,la+lb+1,a,b);
  if ((la < klim) && (lb < klim)) xz(mul_n2)(a,b,c); else xz(mul_k)(a,b,c);
  return((value) c);
}

void xx(mul_k_in)(value x, entier *a, entier *b) {
  entier *c;
  longueur la = Lg(a), lb = Lg(b);
  Enlarge_1_2(c,x,la+lb+1,a,b);
  if ((la < klim) && (lb < klim)) xz(mul_n2)(a,b,c); else xz(mul_k)(a,b,c);
}


                                /* +---------+
                                   |  Carr  |
                                   +---------+ */

value xx(sqr_k)(entier *a) {
  entier *c;
  longueur la = Lg(a);

  Alloc_1_1(c,2*la,a);
  xz(sqr_k)(a,c);
  return((value) c);
}

void xx(sqr_k_in)(value x, entier *a) {
  entier *c;
  longueur la = Lg(a);

  Enlarge_1_1(c,x,2*la,a);
  xz(sqr_k)(a,c);
}

  
                              /* +------------+
                                 |  Division  |
                                 +------------+ */


value xx(quo_1)(entier *a, value b) {
  entier   *q;
  value    res;
  zdouble  bb = Long_val(b);
  zdouble  r;
  longueur la = Lg(a);

  AllocN_1_1(res,2,q,max(la,2),a);
  r = xz(quo_2)(a,bb,q);
  Field(res,0) = (value)q;
  Field(res,1) = Val_long(r);
  return(res);
}

value xx(quo_1_in)(value x,entier *a, value b) {
  entier   *q;
  zdouble  bb = Long_val(b);
  zdouble  r;
  longueur la = Lg(a);

  Enlarge_1_1(q,x,max(la,2),a);
  r = xz(quo_2)(a,bb,q);
  return(Val_long(r));
}

value xx(quo_k)(entier *a, entier *b) {
  entier   *q, *r;
  value    res;
  longueur la = Lg(a), lb = Lg(b), lq, lr;

  lq = (lb > 2) ? max(la-lb+1,1) : max(la,1);
  lr = max(la+1,lb);
  AllocN_2_2(res,2,q,lq,r,max(la+1,lb),a,b);
  if (la < 2*klim) xz(quo_n2)(a,b,q,r); else xz(quo_k)(a,b,q,r);
  Field(res,0) = (value) q;
  Field(res,1) = (value) r;
  return(res);
}


void xx(quo_k_in)(value x, value y, entier *a, entier *b) {
  entier   *q, *r;
  longueur la = Lg(a), lb = Lg(b), lq,lr;

  if (x == y)  failwith("remainder and quotient share the same memory");
  lq = (lb > 2) ? max(la-lb+1,1) : max(la,1);
  lr = max(la+1,lb);
  Enlarge_2_2(q,x,lq,r,y,lr,a,b);
  if (la < 2*klim) xz(quo_n2)(a,b,q,r); else xz(quo_k)(a,b,q,r);
}

value xx(quo_knr)(entier *a, entier *b) {
  entier   *q;
  longueur la = Lg(a), lb = Lg(b), lq;

  lq = (lb > 2) ? max(la-lb+1,1) : max(la,1);
  Alloc_1_2(q,lq,a,b);
  if (la < 2*klim) xz(quo_n2)(a,b,q,NULL); else xz(quo_k)(a,b,q,NULL);
  return((value)q);
}


void xx(quo_knr_in)(value x, entier *a, entier *b) {
  entier   *q;
  longueur la = Lg(a), lb = Lg(b), lq;

  lq = (lb > 2) ? max(la-lb+1,1) : max(la,1);
  Enlarge_1_2(q,x,lq,a,b);
  if (la < 2*klim) xz(quo_n2)(a,b,q,NULL); else xz(quo_k)(a,b,q,NULL);
}

                            /* +-----------------+
                               |  Racine carre  |
                               +-----------------+ */

value xx(sqrt_k)(entier *a) {
  entier *c;
  longueur la = Lg(a), lc=(la+1)/2;

  Alloc_1_1(c,lc+1,a);
  xz(sqrt_k)(a,c);
  return((value) c);
}

void xx(sqrt_k_in)(value x, entier *a) {
  entier *c;
  longueur la = Lg(a), lc=(la+1)/2;

  Enlarge_1_1(c,x,lc+1,a);
  xz(sqrt_k)(a,c);
}

                           /* +--------+
                              |  PGCD  |
                              +--------+ */

value xx(cfrac)(entier *a, entier *b) {
  entier *p,*q,*u,*v,*d;
  value res;
  longueur la = Lg(a), lb = Lg(b), l = max(la,lb)+2;

  AllocN_5_2(res,5,p,l,q,l,u,l,v,l,d,l,a,b);
  Field(res,0) = (value)p;
  Field(res,1) = (value)q;
  Field(res,2) = (value)u;
  Field(res,3) = (value)v;
  Field(res,4) = (value)d;
  xz(cfrac_k)(a,b,p,q,u,v,d);
  return(res);
}

value xx(gcd_ex)(entier *a, entier *b) {
  entier *u,*v,*d;
  value res;
  longueur la = Lg(a), lb = Lg(b), l = max(la,lb)+2;

  AllocN_3_2(res,3,u,l,v,l,d,l,a,b);
  Field(res,0) = (value)u;
  Field(res,1) = (value)v;
  Field(res,2) = (value)d;
  xz(cfrac_k)(a,b,NULL,NULL,u,v,d);
  return(res);
}

value xx(gcd)(entier *a, entier *b) {
  entier *d;
  longueur la = Lg(a), lb = Lg(b), l = max(la,lb)+2;

  Alloc_1_2(d,l,a,b);
  xz(cfrac_k)(a,b,NULL,NULL,NULL,NULL,d);
  return((value)d);
}

void xx(cfrac_in)(value r, value s, value x, value y, value z, entier *a, entier *b) {
  entier *p,*q,*u,*v,*d;
  longueur la = Lg(a), lb = Lg(b), l = max(la,lb)+2;

  if ((r == s) || (r == x) || (r == y) || (r == z) ||
                  (s == x) || (s == y) || (s == z) ||
                              (x == y) || (x == z) ||
                                          (x == z))
    failwith("result sharing with cfrac_in");
  Enlarge_5_2(p,r,l,q,s,l,u,x,l,v,y,l,d,z,l,a,b);
  xz(cfrac_k)(a,b,p,q,u,v,d);
}

void xx(gcd_ex_in)(value x, value y, value z, entier *a, entier *b) {
  entier *u,*v,*d;
  longueur la = Lg(a), lb = Lg(b), l = max(la,lb)+2;

  if ((x == y) || (x == z) || (x == z))
    failwith("result sharing with gcd_ex_in");
  Enlarge_3_2(u,x,l,v,y,l,d,z,l,a,b);
  xz(cfrac_k)(a,b,NULL,NULL,u,v,d);
}

void xx(gcd_in)(value z, entier *a, entier *b) {
  entier *d;
  longueur la = Lg(a), lb = Lg(b), l = max(la,lb)+2;

  Enlarge_1_2(d,z,l,a,b);
  xz(cfrac_k)(a,b,NULL,NULL,NULL,NULL,d);
}

void xx(cfrac_in_bytecode)(value *argv, int argn) {
  xx(cfrac_in)(argv[0],argv[1],argv[2],argv[3],argv[4],(entier *)argv[5],(entier *)argv[6]);
}

                /* +-------------------------------+
                   |  Exponentiation, Factorielle  |
                   +-------------------------------+ */

value xx(pow_k)(entier *a, value b) {
  entier *c;
  longueur lc;
  long bb = Long_val(b);

  if (bb < 0) failwith("negative exponent");
  lc = xz(size_pow_k)(a,bb);
  if (lc < 0) failwith("create too big a number");
  Alloc_1_1(c,lc+1,a);
  xz(pow_k)(a,bb,c,lc);
  return((value) c);
}

void xx(pow_k_in)(value x, entier *a, value b) {
  entier *c;
  longueur lc;
  long bb = Long_val(b);

  if (bb < 0) failwith("negative exponent");
  lc = xz(size_pow_k)(a,bb);
  if (lc < 0) failwith("create too big a number");
  Enlarge_1_1(c,x,lc+1,a);
  xz(pow_k)(a,bb,c,lc);
}

value xx(powmod)(entier *a, entier *b, entier *c) {
  entier *r;

  if (Signe(b))   failwith("negative exponent");
  if (Lg(c) == 0) failwith("division by zero");
  Alloc_1_3(r,Lg(c),a,b,c);
  xz(powmod)(a,b,c,r);
  return((value) r);
}

void xx(powmod_in)(value x, entier *a, entier *b, entier *c) {
  entier *r;

  if (Signe(b) < 0) failwith("negative exponent");
  if (Lg(c) == 0)   failwith("division by zero");
  Enlarge_1_3(r,x,Lg(c),a,b,c);
  xz(powmod)(a,b,c,r);
}

value xx(fact_k)(value a) {
  entier *b;
  long aa = Long_val(a);
  longueur lb = xz(size_fact_k)(aa);

  if (lb < 0) failwith("create too big a number");
  Alloc(b,lb+1);
  xz(fact_k)(aa,b,lb);
  return((value) b);
}

void xx(fact_k_in)(value x, value a) {
  entier *b;
  long aa = Long_val(a);
  longueur lb = xz(size_fact_k)(aa);

  if (lb < 0) failwith("create too big a number");
  Enlarge(b,x,lb+1);
  xz(fact_k)(aa,b,lb);
}

   

                   /* +------------------------+
                      |  Conversion en chaine  |
                      +------------------------+ */

value xx(of_string)(char *s) {
  entier *a;
  longueur la = xz(size_of_string)(s);
  Alloc_1_1t(a,la,char *,s);
  xz(of_string)(s,a);
  return((value)a);
}

value xx(string_of)(entier *a) {
  char *s = xz(string_of)(a);
  value res = copy_string(s);
  free(s);
  return(res);
}

